==> wincdarb_1.0.0_src.tar.gz: wincdarb_1.0.0_src/src/winipcsrv.cpp <==
/*
    WinCdArb - Windows CdRom Device Arbiter service

    Copyright (C) 2007-2019 GuinpinSoft inc <wincdarb@makemkv.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include "winntapi.h"
#include "ntipc.h"
#include <wincdarb/winipc.h>

static LONG volatile g_terminate = 0;

static void InterlockedSet(LONG volatile * _Value,LONG NewValue)
{
    LONG old;
    do
    {
        old = *_Value;
    } while (old!=InterlockedCompareExchange(_Value,NewValue,old));
}

void WINAPI WinIpcClearExitServerLoopFlag()
{
    InterlockedSet(&g_terminate,0);
}

void WINAPI WinIpcSetExitServerLoopFlag()
{
    InterlockedIncrement(&g_terminate);
}

NTSTATUS WINAPI WinIpcCreatePort(PHANDLE PortHandle,PCWSTR PortName)
{
    UNICODE_STRING ustrName;
    OBJECT_ATTRIBUTES oa;

    RtlInitUnicodeString(&ustrName,(PWSTR)PortName);
    InitializeObjectAttributes(&oa,&ustrName,0,NULL,NULL);

    return NtCreatePort(PortHandle,&oa,IPC_MAX_DATA_LENGTH,IPC_MAX_DATA_LENGTH,NULL);
}

NTSTATUS WINAPI WinIpcServerLoop(HANDLE PortHandle,WinIpcProcessMessageRoutine ProcessMessage,PVOID Context)
{
    NTSTATUS        status;
    BOOLEAN         r;
    ULONG           headerLength;
    ULONG           msgSize;
    HANDLE          hConnectHandle;
    LARGE_INTEGER   timeout;
    PUCHAR          msgBuf;
    struct _msg
    {
        IPC_MESSAGE hdr;
        UCHAR       buffer[(IPC_MAX_HEADER_LENGTH+IPC_MAX_DATA_LENGTH)-sizeof(IPC_MESSAGE)];
        UCHAR       justincasenotneededreally[4];
    } msg;

    while (InterlockedIncrement(&g_terminate)<=1)
    {
        InterlockedDecrement(&g_terminate);

        timeout.QuadPart = ((ULONGLONG)0) - 5000*10000ll; // 5000 ms

        status = NtReplyWaitReceivePortEx(
            PortHandle,
            NULL,
            NULL,
            &msg.hdr,
            &timeout);

        if (STATUS_TIMEOUT==status) continue;
        if (STATUS_SUCCESS!=status) return status;

        if (msg.hdr.MessageType!=IPC_MESSAGE_TYPE_CONNECTION_REQUEST)
        {
            continue;
        }

        // we can get 64-bit msg in wow32 process
        headerLength = msg.hdr.Length - msg.hdr.DataLength;
        if ((headerLength!=IPC_HEADER_LENGTH_32) &&
            (headerLength!=IPC_HEADER_LENGTH_64))
        {
            return STATUS_INVALID_PARAMETER;
        }

        // these could not happen under any circumstances
        if (msg.hdr.DataLength>IPC_MAX_DATA_LENGTH) return STATUS_STACK_OVERFLOW;
        if (msg.hdr.Length>(headerLength+IPC_MAX_DATA_LENGTH)) return STATUS_STACK_OVERFLOW;
        if (IPC_HEADER_LENGTH_64==headerLength)
        {
            if (msg.hdr.UniqueProcessIdHigh64!=0) return STATUS_INVALID_HANDLE;
        }

        msgBuf = ((PUCHAR)&msg.hdr)+headerLength;
        msgSize = msg.hdr.DataLength;

        // Avoid data leak. This IS important.
        memset(msgBuf+msgSize,0,IPC_MAX_DATA_LENGTH-msgSize);

        r = (*ProcessMessage)(
            Context,
            msg.hdr.UniqueProcessId,
            msgBuf,
            &msgSize);

        if (FALSE==r)
        {
            msgSize = 0;
        }
        msg.hdr.Length = (USHORT)(headerLength + msgSize);
        msg.hdr.DataLength = (USHORT)msgSize;

        status = NtAcceptConnectPort(&hConnectHandle,NULL,&msg.hdr,r,NULL,NULL);

        if ((STATUS_SUCCESS==status) && (FALSE!=r))
        {
            // Existense of NtReplyWaitReceivePortEx implies that NtCompleteConnectPort is a noop.
            // Keep it just in case, to be on a safe side...
            NtCompleteConnectPort(hConnectHandle);
            NtClose(hConnectHandle);
        }
    }
    return STATUS_SUCCESS;
}

==> wincdarb_1.2.0_src.tar.gz: wincdarb_1.2.0_src/src/winipcsrv.cpp <==
/*
    WinCdArb - Windows CdRom Device Arbiter service

    Copyright (C) 2007-2021 GuinpinSoft inc <wincdarb@makemkv.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include "winntapi.h"
#include "ntipc.h"
#include <wincdarb/winipc.h>
#include <AclAPI.h>

static LONG volatile g_terminate = 0;

static void InterlockedSet(LONG volatile * _Value,LONG NewValue)
{
    LONG old;
    do
    {
        old = *_Value;
    } while (old!=InterlockedCompareExchange(_Value,NewValue,old));
}

void WINAPI WinIpcClearExitServerLoopFlag()
{
    InterlockedSet(&g_terminate,0);
}

void WINAPI WinIpcSetExitServerLoopFlag()
{
    InterlockedIncrement(&g_terminate);
}

typedef struct _WIN_IPC_PORT_SD
{
    SECURITY_DESCRIPTOR      sd;
    EXPLICIT_ACCESSW         ea[1];
    PACL                     acl;
    PSID                     sidWorld;
    SID_IDENTIFIER_AUTHORITY siaWorld;
} WIN_IPC_PORT_SD,*PWIN_IPC_PORT_SD;

static void CleanupPortSD(PWIN_IPC_PORT_SD pSD)
{
    if (NULL != pSD->sidWorld) FreeSid(pSD->sidWorld);
    if (NULL != pSD->acl) LocalFree(pSD->acl);
}

// Everyone: Read Attributes,Connect
static NTSTATUS InitializePortSD(PWIN_IPC_PORT_SD pSD)
{
    DWORD err;

    RtlZeroMemory(pSD, sizeof(*pSD));

    if (!InitializeSecurityDescriptor(&pSD->sd, SECURITY_DESCRIPTOR_REVISION))
    {
        return STATUS_INVALID_PARAMETER;
    }

    pSD->siaWorld = SECURITY_WORLD_SID_AUTHORITY;

    if (!AllocateAndInitializeSid(&pSD->siaWorld, 1, SECURITY_WORLD_RID,
        0, 0, 0, 0, 0, 0, 0,
        &pSD->sidWorld))
    {
        CleanupPortSD(pSD);
        return HRESULT_FROM_WIN32(GetLastError());
    }

    static const ULONG NTIPC_CONNECT_PORT = 1;

    pSD->ea[0].grfAccessPermissions = STANDARD_RIGHTS_READ | NTIPC_CONNECT_PORT;
    pSD->ea[0].grfAccessMode = SET_ACCESS;
    pSD->ea[0].grfInheritance = NO_INHERITANCE;
    pSD->ea[0].Trustee.TrusteeForm = TRUSTEE_IS_SID;
    pSD->ea[0].Trustee.TrusteeType = TRUSTEE_IS_WELL_KNOWN_GROUP;
    pSD->ea[0].Trustee.ptstrName = (LPWSTR)pSD->sidWorld;

    if (ERROR_SUCCESS != (err=SetEntriesInAclW(1, pSD->ea, NULL, &pSD->acl)))
    {
        CleanupPortSD(pSD);
        return HRESULT_FROM_WIN32(err);
    }

    if (!SetSecurityDescriptorDacl(&pSD->sd, TRUE, pSD->acl, FALSE))
    {
        CleanupPortSD(pSD);
        return HRESULT_FROM_WIN32(err);
    }

    return STATUS_SUCCESS;
}

NTSTATUS WINAPI WinIpcCreatePort(PHANDLE PortHandle,PCWSTR PortName)
{
    UNICODE_STRING ustrName;
    OBJECT_ATTRIBUTES oa;
    WIN_IPC_PORT_SD sd;
    NTSTATUS st;

    if (STATUS_SUCCESS != (st = InitializePortSD(&sd))) return st;

    RtlInitUnicodeString(&ustrName,(PWSTR)PortName);
    InitializeObjectAttributes(&oa,&ustrName,0,NULL,&sd.sd);

    st = NtCreatePort(PortHandle,&oa,IPC_MAX_DATA_LENGTH,IPC_MAX_DATA_LENGTH,NULL);

    CleanupPortSD(&sd);

    return st;
}

NTSTATUS WINAPI WinIpcServerLoop(HANDLE PortHandle,WinIpcProcessMessageRoutine ProcessMessage,PVOID Context)
{
    NTSTATUS        status;
    BOOLEAN         r;
    ULONG           headerLength;
    ULONG           msgSize;
    HANDLE          hConnectHandle;
    LARGE_INTEGER   timeout;
    PUCHAR          msgBuf;
    struct _msg
    {
        IPC_MESSAGE hdr;
        UCHAR       buffer[(IPC_MAX_HEADER_LENGTH+IPC_MAX_DATA_LENGTH)-sizeof(IPC_MESSAGE)];
        UCHAR       justincasenotneededreally[4];
    } msg;

    while (InterlockedIncrement(&g_terminate)<=1)
    {
        InterlockedDecrement(&g_terminate);

        timeout.QuadPart = ((ULONGLONG)0) - 5000*10000ll; // 5000 ms

        status = NtReplyWaitReceivePortEx(
            PortHandle,
            NULL,
            NULL,
            &msg.hdr,
            &timeout);

        if (STATUS_TIMEOUT==status) continue;
        if (STATUS_SUCCESS!=status) return status;

        if (msg.hdr.MessageType!=IPC_MESSAGE_TYPE_CONNECTION_REQUEST)
        {
            continue;
        }

        // we can get 64-bit msg in wow32 process
        headerLength = msg.hdr.Length - msg.hdr.DataLength;
        if ((headerLength!=IPC_HEADER_LENGTH_32) &&
            (headerLength!=IPC_HEADER_LENGTH_64))
        {
            return STATUS_INVALID_PARAMETER;
        }

        // these could not happen under any circumstances
        if (msg.hdr.DataLength>IPC_MAX_DATA_LENGTH) return STATUS_STACK_OVERFLOW;
        if (msg.hdr.Length>(headerLength+IPC_MAX_DATA_LENGTH)) return STATUS_STACK_OVERFLOW;
        if (IPC_HEADER_LENGTH_64==headerLength)
        {
            if (msg.hdr.UniqueProcessIdHigh64!=0) return STATUS_INVALID_HANDLE;
        }

        msgBuf = ((PUCHAR)&msg.hdr)+headerLength;
        msgSize = msg.hdr.DataLength;

        // Avoid data leak. This IS important.
        memset(msgBuf+msgSize,0,IPC_MAX_DATA_LENGTH-msgSize);

        r = (*ProcessMessage)(
            Context,
            msg.hdr.UniqueProcessId,
            msgBuf,
            &msgSize);

        if (FALSE==r)
        {
            msgSize = 0;
        }
        msg.hdr.Length = (USHORT)(headerLength + msgSize);
        msg.hdr.DataLength = (USHORT)msgSize;

        status = NtAcceptConnectPort(&hConnectHandle,NULL,&msg.hdr,r,NULL,NULL);

        if ((STATUS_SUCCESS==status) && (FALSE!=r))
        {
            // Existense of NtReplyWaitReceivePortEx implies that NtCompleteConnectPort is a noop.
            // Keep it just in case, to be on a safe side...
            NtCompleteConnectPort(hConnectHandle);
            NtClose(hConnectHandle);
        }
    }
    return STATUS_SUCCESS;
}
